from TISControlProtocol.BytesHelper import checkCRC


# PacketExtractor.py
class PacketExtractor:
    @staticmethod
    def extract_info(packet: list):
        packet_check = checkCRC(packet)
        info = {}
        if packet_check:
            print("correct packet")
            info["source_ip"] = packet[0:4]
            info["device_id"] = packet[17:19]
            info["device_type"] = packet[19:21]
            info["operation_code"] = packet[21:23]
            info["additional_bytes"] = packet[25:-2]

        else:
            print("wrong packet")
        return info



from socket import socket, SOL_SOCKET, SO_BROADCAST

from TISControlProtocol.Protocols.udp.AckCoordinator import AckCoordinator
import asyncio
from abc import ABC, abstractmethod
from TISControlProtocol.shared import ack_events  # noqa: F401
from collections import deque
import logging


# PacketSender.py
class PacketSender:
    def __init__(self, socket: socket, coordinator: AckCoordinator, UDP_IP, UDP_PORT):
        self.UDP_IP = UDP_IP
        self.UDP_PORT = UDP_PORT
        self.socket = socket
        self.socket.setsockopt(SOL_SOCKET, SO_BROADCAST, 1)
        self.coordinator = coordinator
        self.command_stacks = {}  # Holds the command stacks
        self.last_command_times = {}  # Holds the last command times for debouncing
        self.update_packet_queue = deque()  # holds update packets
        self.update_device_queue = set()  # holds update device ids

    async def send_packet(self, packet: list, destination: str):
        print(f"sending {packet}")
        self.socket.sendto(bytes(packet), (destination, self.UDP_PORT))

    async def send_packet_with_ack(
        self,
        destination: str,
        packet: list,
        packet_dict: dict = None,
        channel_number: int = None,
        attempts: int = 10,
        timeout: float = 0.5,
        debounce_time: float = 0.5,  # The debounce time in seconds
    ):
        unique_id = (
            tuple(packet_dict["device_id"]),
            tuple(packet_dict["operation_code"]),
            channel_number,
        )

        # Add the command to the stack for this unique ID
        if unique_id not in self.command_stacks:
            self.command_stacks[unique_id] = []
        self.command_stacks[unique_id].append(packet)

        # Only process the last command in the stack
        if packet != self.command_stacks[unique_id][-1]:
            return

        # If the command is being called too quickly after the last one, ignore it
        if (
            unique_id in self.last_command_times
            and asyncio.get_event_loop().time() - self.last_command_times[unique_id]
            < debounce_time
        ):
            return

        self.last_command_times[unique_id] = asyncio.get_event_loop().time()

        event = self.coordinator.create_ack_event(unique_id)

        for attempt in range(attempts):
            await self.send_packet(packet, destination)
            try:
                await asyncio.wait_for(event.wait(), timeout)
                logging.error(f"ack received for {unique_id}")
                # Remove the command from the stack after it's processed
                self.command_stacks[unique_id].remove(packet)
                return True
            except asyncio.TimeoutError:
                print(
                    f"ack not received within {timeout} seconds, attempt {attempt + 1}"
                )
                logging.error(f"ack not received within {timeout} seconds")

        self.coordinator.remove_ack_event(unique_id)
        logging.error(f"ack not received after {attempts} attempts")
        return False

    async def brodcast_packet(self, packet: list):
        print(f"broadcasting {packet}")
        self.socket.sendto(bytes(packet), ("<broadcast>", self.UDP_PORT))

    # # update
    # def queue_update_packet(self,device_id:list,):
    #     if entity._device_id not in



from asyncio import get_event_loop, AbstractEventLoop
import socket
from TISControlProtocol.Protocols.udp.PacketProtocol import PacketProtocol

loop = get_event_loop()


async def setup_udp_protocol(
    sock: socket, loop: AbstractEventLoop, udp_ip, udp_port, local_ip, hass
) -> tuple[socket.socket, PacketProtocol]:
    transport, protocol = await loop.create_datagram_endpoint(
        lambda: PacketProtocol(sock, udp_ip, udp_port, hass),
        remote_addr=(udp_ip, udp_port),
        local_addr=("0.0.0.0", udp_port),
        allow_broadcast=True,
        reuse_port=True,
    )
    return transport, protocol



from TISControlProtocol.Protocols import setup_udp_protocol
from TISControlProtocol.BytesHelper import build_packet
from homeassistant.core import HomeAssistant  # type: ignore
from homeassistant.components.http import HomeAssistantView  # type: ignore
from homeassistant.components import homeassistant  # type: ignore
from typing import Optional
from aiohttp import web  # type: ignore
import socket
import logging
from collections import defaultdict
import json
import asyncio
import ST7789
from PIL import Image, ImageDraw, ImageFont  # noqa: F401


class TISApi:
    """TIS API class."""

    def __init__(
        self,
        host: str,
        port: int,
        local_ip: str,
        hass: HomeAssistant,
        domain: str,
        devices_dict: dict,
        display_logo: Optional[str] = "./custom_components/tis_control/shakalpng.png",
    ):
        """Initialize the API class."""
        self._host = host
        self._port = port
        self._local_ip = local_ip
        self._protocol = None
        self._transport = None
        self._hass = hass
        self._config_entries = {}
        self._domain = domain
        self._devices_dict = devices_dict
        self._display_logo = display_logo
        self._display = None

    async def connect(self):
        """Connect to the TIS API."""
        self.loop = self._hass.loop
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            self._transport, self._protocol = await setup_udp_protocol(
                self._sock,
                self.loop,
                self._host,
                self._port,
                self._local_ip,
                self._hass,
            )
            logging.error(
                f"Connected to TIS API successfully, ip_comport:{self._host},local:{self._local_ip}"
            )
        except Exception as e:
            logging.error(f"Error connecting to TIS API: {e}")

    def run_display(self, style="dots"):
        try:
            self.display = ST7789.ST7789(
                width=320,
                height=240,
                rotation=0,
                port=0,
                cs=0,
                dc=23,
                rst=25,
                backlight=12,
                spi_speed_hz=60 * 1000 * 1000,
                offset_left=0,
                offset_top=0,
            )
            # Initialize display.
            self.display.begin()
            self.set_display_image()

        except Exception as e:
            logging.error(f"error initializing display, {e}")
            return

    def set_display_image(self):
        img = Image.open(self._display_logo)
        self.display.set_backlight(0)
        # reset display
        self.display.display(img)

    async def _parse_device_manager_request(self, data: dict) -> None:
        """Parse the device manager request."""
        converted = {
            appliance: {
                "device_id": [int(n) for n in details[0]["device_id"].split(",")],
                "appliance_type": details[0]["appliance_type"]
                .lower()
                .replace(" ", "_"),
                "appliance_class": details[0].get("appliance_class", None),
                "is_protected": bool(int(details[0]["is_protected"])),
                "gateway": details[0]["gateway"],
                "channels": [
                    {
                        "channel_number": int(detail["channel_number"]),
                        "channel_type": detail["channel_type"],
                        "channel_name": detail["channel_name"],
                    }
                    for detail in details
                ],
            }
            for appliance, details in data["appliances"].items()
        }

        grouped = defaultdict(list)
        for appliance, details in converted.items():
            if (
                details["appliance_type"]
                in self._hass.data[self._domain]["supported_platforms"]
            ):
                logging.error(f"appliance: {appliance}, details: {details}")
                grouped[details["appliance_type"]].append({appliance: details})

        self._config_entries = dict(grouped)
        # add a lock module config entry
        self._config_entries["lock_module"] = {
            "password": data["configs"]["lock_module_password"]
        }
        logging.error(f"config_entries stored: {self._config_entries}")
        # await self.update_entities()

    async def get_entities(self, platform: str = None):
        """Get the entities."""
        try:
            with open("appliance_data.json", "r") as f:
                data = json.load(f)
                await self._parse_device_manager_request(data)
                logging.error(f"Data loaded from file: {data}")
        except Exception as e:
            # handle or log the exception
            logging.error(f"Error loading data from file: {e},file: {data}")
        await self._parse_device_manager_request(data)
        return self._config_entries[platform]


class TISEndPoint(HomeAssistantView):
    """TIS API endpoint."""

    url = "/api/tis"
    name = "api:tis"
    requires_auth = False

    def __init__(self, tis_api: TISApi):
        """Initialize the API endpoint."""
        self.api = tis_api

    async def post(self, request):
        # Parse the JSON data from the request
        data = await request.json()
        # dump to file
        with open("appliance_data.json", "w") as f:
            json.dump(data, f, indent=4)
        # reload the platforms
        for entry in self.api._hass.config_entries.async_entries(self.api._domain):
            await self.api._hass.config_entries.async_reload(entry.entry_id)
        await self.api._hass.services.async_call(
            self.api._domain, homeassistant.SERVICE_RELOAD_ALL
        )


class ScanDevicesEndPoint(HomeAssistantView):
    """Scan Devices API endpoint."""

    url = "/api/scan_devices"
    name = "api:scan_devices"
    requires_auth = False

    def __init__(self, tis_api: TISApi):
        """Initialize the API endpoint."""
        self.api = tis_api
        self._discovery_packet = build_packet(
            operation_code=[0x00, 0x0E],
            ip_address=self.api._host,
            destination_mac="FF:FF:FF:FF:FF:FF",
            device_id=[0xFF, 0xFF],
            additional_packets=[],
        )

    async def get(self, request):
        # Discover network devices
        devices = await self.discover_network_devices()
        devices = [
            {
                "device_id": device["device_id"],
                "device_type": self.api._devices_dict.get(
                    tuple(device["device_type"]), tuple(device["device_type"])
                ),
                "gateway": device["source_ip"],
            }
            for device in devices
        ]
        # TODO: some processing and formating
        return web.json_response(devices)

    async def discover_network_devices(self, prodcast_attempts=10) -> list:
        # empty current discovered devices list
        self.api._hass.data[self.api._domain]["discovered_devices"] = []
        for i in range(prodcast_attempts):
            await self.api._protocol.sender.brodcast_packet(self._discovery_packet)
            # sleep for 1 sec
            await asyncio.sleep(1)

        return self.api._hass.data[self.api._domain]["discovered_devices"]



from TISControlProtocol.BytesHelper import *  # noqa: F403
from TISControlProtocol.Protocols.udp.PacketSender import PacketSender
from TISControlProtocol.Protocols.udp.PacketReceiver import PacketReceiver
from TISControlProtocol.Protocols.udp.AckCoordinator import AckCoordinator

from TISControlProtocol.shared import ack_events

from homeassistant.core import HomeAssistant  # type: ignore
from .PacketHandlers.BinaryFeedbackHandler import handle_binary_feedback
from .PacketHandlers.ControlResponseHandler import handle_control_response

from .PacketHandlers.AutoBinaryFeedbackHandler import handle_auto_binary_feedback

from .PacketHandlers.ClimateControlFeedbackHandler import (
    handle_climate_control_feedback,
)
from .PacketHandlers.ClimateBinaryFeedbackHandler import handle_climate_binary_feedback
from .PacketHandlers.DiscoveryFeedbackHandler import handle_discovery_feedback
from .PacketHandlers.UpdateResponseHandler import handle_update_response

import socket as Socket

OPERATIONS_DICT = {
    (0x00, 0x32): handle_control_response,
    (0xEF, 0xFF): handle_binary_feedback,
    (0xDC, 0x22): handle_auto_binary_feedback,
    (0xE0, 0xEF): handle_climate_control_feedback,
    (0xE3, 0xD9): handle_climate_binary_feedback,
    (0x00, 0x0F): handle_discovery_feedback,
    (0x00, 0x34): handle_update_response,
}
# 1C 01 30 1B BA DC 22 FF FF 08 02 02 02 02 02 02 02 02 00 01 01 01 01 01 01 01 57 62


class PacketProtocol:
    def __init__(
        self,
        socket: Socket.socket,
        UDP_IP,
        UDP_PORT,
        hass: HomeAssistant,
    ):
        self.UDP_IP = UDP_IP
        self.UDP_PORT = UDP_PORT
        self.socket = socket
        self.searching = False
        self.search_results = []
        self.discovered_devices = []
        self.hass = hass

        self.ack_events = ack_events
        self.coordinator = AckCoordinator()
        self.sender = PacketSender(
            socket=self.socket,
            coordinator=self.coordinator,
            UDP_IP=self.UDP_IP,
            UDP_PORT=self.UDP_PORT,
        )
        self.receiver = PacketReceiver(self.socket, OPERATIONS_DICT, self.hass)

        self.connection_made = self.receiver.connection_made
        self.datagram_received = self.receiver.datagram_received



from TISControlProtocol.shared import ack_events
import asyncio
from homeassistant.core import HomeAssistant

import logging


async def handle_climate_control_feedback(hass: HomeAssistant, info: dict):
    ac_number = info["additional_bytes"][1]
    state = info["additional_bytes"][2]
    cool_temp = info["additional_bytes"][3]
    hvac_mode = (info["additional_bytes"][4] >> 4) & 0x0F
    fan_speed = info["additional_bytes"][4] & 0x0F
    heat_temp = info["additional_bytes"][7]
    auto_temp = info["additional_bytes"][9]
    dry_temp = info["additional_bytes"][10]

    event_data = {
        "device_id": info["device_id"],
        "ac_number": ac_number,
        "state": state,
        "cool_temp": cool_temp,
        "hvac_mode": hvac_mode,
        "fan_speed": fan_speed,
        "heat_temp": heat_temp,
        "auto_temp": auto_temp,
        "dry_temp": dry_temp,
    }
    logging.error(f"climate control feedback: {event_data}")

    try:
        hass.bus.async_fire(str(info["device_id"]), event_data)
    except Exception as e:
        logging.error(f"error in firing even for feedbackt: {e}")

    try:
        event: asyncio.Event = ack_events.get(
            (
                tuple(info["device_id"]),
                (0xE0, 0xEE),
                ac_number,
            )
        )
        if event is not None:
            print("setting event")
            event.set()
    except Exception as e:
        print(e)



from homeassistant.core import HomeAssistant  # type: ignore
import logging


async def handle_update_response(hass: HomeAssistant, info: dict):
    # remove auxilary bytes which represents number of scenarios
    channels_number = info["additional_bytes"][0]
    event_data = {
        "device_id": info["device_id"],
        "feedback_type": "update_response",
        "additional_bytes": info["additional_bytes"],
    }
    try:
        hass.bus.async_fire(str(info["device_id"]), event_data)
        logging.error(f"update response event fired for {info['device_id']}")
    except Exception as e:
        logging.error(f"error in firing event: {e}")



# deprecated
async def handle_search_response(self, info: dict):
    print(f"got search response packet from {info['device_id']}")
    self.discovered_devices.append(
        {
            "device_id": info["device_id"],
            "device_type": DEVICES_DICT[  # noqa: F821  # type: ignore
                tuple(info["additional_bytes"])
            ],
            "gateway": info["source_ip"],
        }
    )



from TISControlProtocol.BytesHelper import *  # noqa: F403
from socket import socket
from TISControlProtocol.Protocols.udp.PacketExtractor import PacketExtractor
from TISControlProtocol.Protocols.udp.PacketDispatcher import PacketDispatcher
import logging
from homeassistant.core import HomeAssistant  # type: ignore


# PacketReceiver.py
class PacketReceiver:
    def __init__(
        self,
        socket: socket,
        OPERATIONS_DICT: dict,
        hass: HomeAssistant,
    ):
        self.socket = socket
        self._hass = hass
        self.dispatcher = PacketDispatcher(self._hass, OPERATIONS_DICT)
        self.transport = None
        self

    def connection_made(self, transport):
        self.transport = transport
        logging.error("connection made")

    def datagram_received(self, data, addr):
        # logging.error(f"datagram received {data} from {addr}")
        try:
            hex = bytes2hex(data, [])  # noqa: F405
            logging.warning(f"recieved packet {hex}")
            info = PacketExtractor.extract_info(hex)
            # dispatch the packet to the appropriate method according to the info
            self._hass.async_create_task(self.dispatcher.dispatch_packet(info))

        except Exception as e:
            logging.error(f"Error in datagram_received: {e}, info: {info}")



from homeassistant.core import HomeAssistant
import logging

AC_NUMBER_MAP = {0x19: 0, 0x1A: 1, 0x1B: 2, 0x1C: 3, 0x1D: 4, 0x1E: 5, 0x1F: 6, 0x20: 7}


async def handle_climate_binary_feedback(hass: HomeAssistant, info: dict):
    ac_number = AC_NUMBER_MAP.get(info["additional_bytes"][0], None)
    if ac_number is None:
        ac_number = 0
        sub_operation = info["additional_bytes"][0]
        operation_value = info["additional_bytes"][1]

    else:
        sub_operation = info["additional_bytes"][1]
        operation_value = info["additional_bytes"][2]

    event_data = {
        "device_id": info["device_id"],
        "ac-number": ac_number,
        "sub_operation": sub_operation,
        "operation_value": operation_value,
    }
    try:
        hass.bus.async_fire(str(info["device_id"]), event_data)
        logging.error(
            f"ac binary feedback event: {event_data} , additional bytes: {info['additional_bytes']}"
        )
    except Exception as e:
        logging.error(f"error in firing event: {e}")
